// ksmstool
// Copyright (C) 2006 by Joachim Gehweiler
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

/****************************************************************************
** ui.h extension file, included from the uic-generated form implementation.
**
** If you want to add, delete, or rename functions or slots, use
** Qt Designer to update this file, preserving your code.
**
** You should not define a constructor or destructor in this file.
** Instead, write your code in functions called init() and destroy().
** These will automatically be called by the form's constructor and
** destructor.
*****************************************************************************/

// a lot of the GSM specific code is borrowed from "SMS Server Tools" which is under GPL, too
// http://smstools.meinemullemaus.de/

#include <qapplication.h>
#include <qdir.h>
#include <qfile.h>
#include <qmap.h>
#include <qmessagebox.h>
#include <qprogressdialog.h>
#include <qregexp.h>
#include <qstring.h>

#include <assert.h>
#include <fcntl.h>
#include <termios.h>
#include <regex.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>


void SMSTool::init()
{
    // read addressbook
    QRegExp regex("^([\\+0-9]+)[\\t,;](.+)$");
    QFile file(QString("%1/.ksmstool/addressbook.csv").arg(QDir::homeDirPath()));
    if(file.open(IO_ReadOnly))
    {
	QTextStream stream(&file);
	QString line;
	while (!stream.eof())
	{
	    line = stream.readLine();
	    if(regex.search(line) == -1)
		QMessageBox::warning(this, "SMS-Tool", QString("Skipping recipient '%1' in wrong format...").arg(line), QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton, QMessageBox::NoButton);
	    else
	    {
		m_numbers.insert(regex.cap(2), regex.cap(1));
		predefinedCombo->insertItem(regex.cap(2));
	    }
	}
	file.close();
    }
    else
    {
	QDir dir = QDir(QString("%1/.ksmstool").arg(QDir::homeDirPath()));
	if(!dir.exists())
	    dir.mkdir(QString("%1/.ksmstool").arg(QDir::homeDirPath()));
	
	QMessageBox::warning(this, "SMS-Tool", QString("Could not load list of predefined recipients from %1/.ksmstool/addressbook.csv!").arg(QDir::homeDirPath()), QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton, QMessageBox::NoButton);
    }
    
    // initialize randoms (for concatenated messages)
    srand(time(NULL));
}

void SMSTool::send_sms()
{
    // get message + recipient
    QString recipient, recipientName, message;
    
    if(recipientGroup->selected() == predefinedRadio)
    {
	recipientName = predefinedCombo->currentText();
	QMap<QString, QString>::iterator it = m_numbers.find(recipientName);
	if(it == m_numbers.end())
	{
	    QMessageBox::critical(this, "SMS-Tool", "Phone number of recipient not found!", QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton, QMessageBox::NoButton);
	    return;
	}
	recipient = it.data();
    }
    else
    {
	recipient = numberEdit->text();
	recipientName = recipient;
    }
    
    message = messageEdit->text();
    
    if(message.isEmpty())
    {
	QMessageBox::critical(this, "SMS-Tool", "Cannot send an empty SMS!", QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton, QMessageBox::NoButton);
	return;
    }
    
    // setup progress dialog
    QProgressDialog progress("Opening GSM modem...", 0, 1100, this, "progress", true);
    progress.setMinimumDuration(0);
    progress.setProgress(1);
    
    // open + init modem
    int modem;
    
    QString modemName = "/dev/ttyS0";
    if(qApp->argc() == 2)
	modemName = qApp->argv()[1];
    
    modem = open(modemName, O_RDWR | O_NOCTTY | O_NONBLOCK);
    if(modem < 0)
    {
	progress.reset();
	QMessageBox::critical(this, "SMS-Tool", QString("Could not open GSM modem %1!").arg(modemName), QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton, QMessageBox::NoButton);
	return;
    }
    
    struct termios newtio;
    bzero(&newtio, sizeof(newtio));
    newtio.c_cflag = CS8 | CLOCAL | CREAD | O_NDELAY | O_NONBLOCK | CRTSCTS;
    newtio.c_iflag = IGNPAR;
    newtio.c_oflag = 0;
    newtio.c_lflag = 0;
    newtio.c_cc[VTIME] = 0;
    newtio.c_cc[VMIN] = 0;
    
    cfsetispeed(&newtio, B115200);
    cfsetospeed(&newtio, B115200);
    
    tcsetattr(modem, TCSANOW, &newtio);
    
    if(!init_modem(modem, progress))
	return;
    
    progress.setProgress(100);
    progress.setLabelText("Converting SMS character set...");

    // convert message into GSM charset
    int l = message.length();
    char *msg = (char*)malloc(2*l+1);
    assert(msg != NULL);
    l = iso2gsm(message.ascii(), l, msg, 2*l);
    
    // split message if necessary
    bool success = true;
    char udh[13];
    
    int parts = 1;
    if(l > 160)
    {
	// cut off 7 characters for the 48-bit user data header (UDH)
	int partlen = 153;
	parts = (l + partlen - 1) / partlen;
    }
    
    if(parts == 1)
    {
	progress.setLabelText("Sending SMS...");

	udh[0] = 0;
	if(!send_part(modem, recipient.ascii(), msg, l, udh, progress))
	{
	    progress.reset();
	    QMessageBox::critical(this, "SMS-Tool", "Failed to send SMS!", QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton, QMessageBox::NoButton);
	    success = false;
	}
    }
    else
    {
	char part[161];
	int i, r, rnd = rand() % 256;
		
	progress.setTotalSteps(parts*1000 + 100);
	
	for(i=0; i<parts; i++)
	{
	    progress.setLabelText(QString("Sending part %1 (out of %2)...").arg(i+1).arg(parts));
	    
	    r = l - (i*153);
	    if(r > 153)
		r = 153;
	    // GSM_03.40_5.3.0.pdf
	    sprintf(udh, "050003%02X%02X%02X", rnd, parts, i+1);
	    memcpy(part, msg+(i*153), r);
	    if(!send_part(modem, recipient.ascii(), part, r, udh, progress))
	    {
		progress.reset();
		if(i == 0)
		    QMessageBox::critical(this, "SMS-Tool", "Failed to send SMS!", QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton, QMessageBox::NoButton);
		else
		    QMessageBox::critical(this, "SMS-Tool", QString("Failure while sending the SMS:\nonly %1 out of %2 parts could be sent!").arg(i).arg(parts), QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton, QMessageBox::NoButton);
		
		success = false;
		break;
	    }
	    
	    progress.setProgress((i+1)*1000 + 100);
	}
    }
    
    if(success)
    {
	progress.setProgress(progress.totalSteps());
	QMessageBox::information(this, "SMS-Tool", "The SMS has been sent!", QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton, QMessageBox::NoButton);
    }
    
    // clean up
    free(msg);
    
    // close does not work for some mysterious reasons...
    // close(modem);
    
    if(!success)
	return;
    
    // save SMS in sent file
    QFile file(QString("%1/.ksmstool/sent.txt").arg(QDir::homeDirPath()));
    if(file.open(IO_WriteOnly | IO_Append))
    {
	QTextStream stream(&file);
	stream << recipientName << "\t" << message << "\n";
	file.close();
    }
    else
	QMessageBox::warning(this, "SMS-Tool", "Failed to write SMS to sent list!", QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton, QMessageBox::NoButton);
}

void SMSTool::set_message_label()
{
    int n = messageEdit->text().length(), parts = 1, r;
    
    if(n > 160)
    {
	// cut off 7 characters for the 48-bit user data header (UDH)
	int partlen = 153;
	parts = (n + partlen - 1) / partlen;
    }
    
    if(parts == 1)
	r = 160 - n;
    else
	r = 153 - n % 153;
    
    messageLabel->setText(QString("Message  (%1 SMS, %2 characters left):").arg(parts).arg(r));
}

// http://www.csoft.co.uk/sms/character_sets/gsm.htm
char SMSTool::charset[] = {
    '@' , 0x00,
    0xA3, 0x01,
    '$' , 0x02,
    0xA5, 0x03,
    0xE8, 0x04,
    0xE9, 0x05,
    0xF9, 0x06,
    0xEC, 0x07,
    0xF2, 0x08,
    0xC7, 0x09,
    '\n', 0x0A,
    0xD8, 0x0B,
    0xF8, 0x0C,
    '\r', 0x0D,
    0xC5, 0x0E,
    0xE5, 0x0F,
    0x81, 0x10,
    '_', 0x11,
    0x82, 0x12,
    0x83, 0x13,
    0x84, 0x14,
    0x85, 0x15,
    0x86, 0x16,
    0x87, 0x17,
    0x88, 0x18,
    0x89, 0x19,
    0x8A, 0x1A,
    0x1B, 0x1B,
    0xC6, 0x1C,
    0xE6, 0x1D,
    0xDF, 0x1E,
    0xC9, 0x1F,
    ' ' , 0x20,
    '!' , 0x21,
    '\"', 0x22,
    '#' , 0x23,
    '%' , 0x25,
    '&' , 0x26,
    '\'', 0x27,
    '(' , 0x28,
    ')' , 0x29,
    '*' , 0x2A,
    '+' , 0x2B,
    ',' , 0x2C,
    '-' , 0x2D,
    '.' , 0x2E,
    '/' , 0x2F,
    '0' , 0x30,
    '1' , 0x31,
    '2' , 0x32,
    '3' , 0x33,
    '4' , 0x34,
    '5' , 0x35,
    '6' , 0x36,
    '7' , 0x37,
    '8' , 0x38,
    '9' , 0x39,
    ':' , 0x3A,
    ';' , 0x3B,
    '<' , 0x3C,
    '=' , 0x3D,
    '>' , 0x3E,
    '?' , 0x3F,
    0xA1, 0x40,
    'A' , 0x41,
    'B' , 0x42,
    'C' , 0x43,
    'D' , 0x44,
    'E' , 0x45,
    'F' , 0x46,
    'G' , 0x47,
    'H' , 0x48,
    'I' , 0x49,
    'J' , 0x4A,
    'K' , 0x4B,
    'L' , 0x4C,
    'M' , 0x4D,
    'N' , 0x4E,
    'O' , 0x4F,
    'P' , 0x50,
    'Q' , 0x51,
    'R' , 0x52,
    'S' , 0x53,
    'T' , 0x54,
    'U' , 0x55,
    'V' , 0x56,
    'W' , 0x57,
    'X' , 0x58,
    'Y' , 0x59,
    'Z' , 0x5A,
    0xC4, 0x5B,
    0xD6, 0x5C,
    0xD1, 0x5D,
    0xDC, 0x5E,
    '', 0x5F,
    0xBF, 0x60,
    'a' , 0x61,
    'b' , 0x62,
    'c' , 0x63,
    'd' , 0x64,
    'e' , 0x65,
    'f' , 0x66,
    'g' , 0x67,
    'h' , 0x68,
    'i' , 0x69,
    'j' , 0x6A,
    'k' , 0x6B,
    'l' , 0x6C,
    'm' , 0x6D,
    'n' , 0x6E,
    'o' , 0x6F,
    'p' , 0x70,
    'q' , 0x71,
    'r' , 0x72,
    's' , 0x73,
    't' , 0x74,
    'u' , 0x75,
    'v' , 0x76,
    'w' , 0x77,
    'x' , 0x78,
    'y' , 0x79,
    'z' , 0x7A,
    0xE4, 0x7B,
    0xF6, 0x7C,
    0xF1, 0x7D,
    0xFC, 0x7E,
    0xE0, 0x7F,
    0, 0 // end marker
};

// extended characters (preceeded by 0x1B)
char SMSTool::ext_charset[] = {
    0x0C, 0x0A,
    '^', 0x14,
    '{' , 0x28,
    '}' , 0x29,
    '\\', 0x2F,
    '[' , 0x3C,
    '~' , 0x3D,
    ']' , 0x3E,
    '|', 0x40,
    0xA4 , 0x65,
    0, 0 // end marker
};

int SMSTool::write_to_modem(int modem, char *command)
{  
    int status = 0, timeoutcounter = 0, x = 0;
    struct termios tio;
    
    tcgetattr(modem, &tio);
    if(tio.c_cflag & CRTSCTS)
    {
	ioctl(modem, TIOCMGET, &status);
	while(!(status & TIOCM_CTS))
	{
	    usleep(100000);
	    timeoutcounter++;
	    ioctl(modem, TIOCMGET, &status);
	    if(timeoutcounter > 30)
		return 0;
	}
    }
    
    for(x=0; x<(int)strlen(command); x++) 
    {
	if(write(modem, command+x, 1) < 1)
	    return 0;
	usleep(1000);
	tcdrain(modem);
    }
    
    return 1;
}

// Read max characters from modem. The function returns when it received at 
// least 1 character and then the modem is quiet for 200ms.
// The answer might contain already a string. In this case, the answer 
// gets appended to this string.
int SMSTool::read_from_modem(int modem, char *answer, int max)
{
    int count = 0, got = 0, timeoutcounter = 0, success = 0, toread = 0;
    
    do 
    {
	// read no more than buffer size - 1
	count = strlen(answer);
	toread = max - count - 1;
	if(toread <= 0)
	    break;
	
	got = read(modem, answer + count, toread);
	// if nothing received ...
	if (got<=0)
	{
	    // wait a litte bit and then repeat this loop
	    got = 0;
	    usleep(100000);
	    timeoutcounter++;
	}
	else  
	{
	    // restart timout counter
	    timeoutcounter = 0;
	    
	    // append a string termination character
	    answer[count+got] = 0;
	    success = 1;
	}
    }
    while (timeoutcounter < 2);
    
    return success;
}

bool SMSTool::init_modem(int modem, QProgressDialog &progress)
{
    char answer[500];
    int retries = 0, success = 0;
    
    // check if modem is ready
    retries=0;
    do
    {
	retries++;
	modem_command(modem, "AT\r", answer, sizeof(answer), 50, "(OK)|(ERROR)", NULL);
	if(strstr(answer, "OK") == NULL && strstr(answer, "ERROR") == NULL)
	{
	    // modem does not answer, try to send a PDU termination character
	    modem_command(modem, "\x1A\r", answer, sizeof(answer), 50, "(OK)|(ERROR)", NULL);
	    answer[0] = 0; //ignore answer
	}
    }
    while((retries <= 10) && (!strstr(answer, "OK")));
    
    if(!strstr(answer, "OK"))
    {
	QMessageBox::critical(this, "SMS-Tool", "GSM modem does not answer!", QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton, QMessageBox::NoButton);
	return false;
    }
    
    progress.setProgress(30);
    progress.setLabelText("Checking for network...");
    
    // check if modem is registered to the network
    success = 0;
    retries = 0;
    do
    {
	retries++;
	modem_command(modem, "AT+CREG?\r", answer, sizeof(answer), 100, "(\\+CREG:.*OK)|(ERROR)", NULL);
	if(strstr(answer, "1")) // registered
	    success=1;
	else if(strstr(answer, "5")) // registered to a roaming partner
	    success=1;
	else if(strstr(answer, "+CREG:")) // not registered, waiting before retrying
	    sleep(3);
	else // error, waiting before retrying
	    sleep(3);
    }
    while((success == 0) && (retries < 10));
    
    if(success == 0)
    {
	QMessageBox::critical(this, "SMS-Tool", "GSM modem is not registered to a network!", QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton, QMessageBox::NoButton);
	return false;
    }
    
    progress.setProgress(70);
    progress.setLabelText("Initializing GSM modem...");
    
    // selecting PDU mode
    modem_command(modem, "AT+CMGF=0\r", answer, sizeof(answer), 50, "(OK)|(ERROR)", NULL);
    if(strstr(answer, "ERROR"))
    {
	QMessageBox::critical(this, "SMS-Tool", "Could not initialize GSM modem!", QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton, QMessageBox::NoButton);
	return false;
    }
    
    return true;
}

int SMSTool::modem_command(int modem, char *command, char *answer, int max, int timeout, char *expect, QProgressDialog *progress)
{
    int timeoutcounter=0, t;
    regex_t re;
    
    assert(expect != NULL);
    assert(expect[0] != 0);
    
    // compile regular expressions
    t = regcomp(&re, expect, REG_EXTENDED | REG_NOSUB);
    assert (t == 0); 
    
    // clean input buffer
    // It seems that this command does not do anything because actually it 
    // does not clear the input buffer. However I do not remove it until I 
    // know why it does not work.
    tcflush(modem, TCIFLUSH);
  
    // send command
    if(write_to_modem(modem, command) == 0)
	return 0;
    
    int offset = 0;
    double step = 0.0;
    
    if(progress != NULL)
    {
	offset = progress->progress();
	step = 800.0 / (double)timeout;
	
	progress->setProgress(offset + 100);
    }
    
    // wait for the answer 
    answer[0]=0;
    timeoutcounter = 0;
    do
    {
	read_from_modem(modem, answer, max);
	
	// check if it is the expected answer
	if(regexec(&re, answer, (size_t)0, NULL, 0) == 0)
	    break;
	timeoutcounter+=2;
	
	if(progress != NULL)
	    progress->setProgress(offset + 100 + (int)(step * (double)timeoutcounter));
    }
    while(timeoutcounter < timeout);
    
    // free memory used by regexp
    regfree(&re);
    
    return strlen(answer);
}

void SMSTool::swapchars(char *string) 
{
    char c;
    int length = strlen(string), position;
    
    for(position=0; position<length-1; position+=2)
    {
	c=string[position];
	string[position]=string[position+1];
	string[position+1]=c;
    }
}

int SMSTool::ascii2pdu(char *ascii, int length, char *pdu)
{
    char converted, tmp[500], octet[10];
    int pdubitposition, pdubyteposition = 0, character, bit, pdubitnr, counted_characters = 0;
    
    assert(length <= 160);
    
    // clear the tmp buffer
    bzero(tmp, sizeof(tmp));
    
    for(character=0; character<length; character++)
    {
	converted = ascii[character];
	counted_characters++;
	for(bit=0; bit<7; bit++)
	{
	    pdubitnr = 7 * character + bit;
	    pdubyteposition = pdubitnr / 8;
	    pdubitposition = pdubitnr % 8;
	    if(converted & (1<<bit))
		tmp[pdubyteposition] = tmp[pdubyteposition] | (1<<pdubitposition);
	    else
		tmp[pdubyteposition] = tmp[pdubyteposition] & ~(1<<pdubitposition);
	}
    }
    tmp[pdubyteposition+1] = 0;
    pdu[0] = 0;
    for(character=0; character<=pdubyteposition; character++)
    {
	sprintf(octet, "%02X", (unsigned char)tmp[character]);
	strcat(pdu,octet);
    }
    return counted_characters;
}

int SMSTool::iso2gsm(const char *source, int size, char *destination, int max)
{
    int table_row, source_count = 0, dest_count = 0, found = 0;
    
    assert(source != NULL);
    assert(size > 0);
    
    destination[dest_count] = 0;
    
    while(source_count < size && dest_count < max)
    {
	// search in normal translation table
	found = 0;
	table_row = 0;
	while(charset[table_row*2])
	{
	    if(charset[table_row*2] == source[source_count])
	    {
		destination[dest_count++] = charset[table_row*2+1];
		found = 1;
		break;
	    }
	    table_row++;
	}
	
	// search in the extended table
	if(found == 0)
	{
	    table_row = 0;
	    while(ext_charset[table_row*2])
	    {
		if(ext_charset[table_row*2] == source[source_count])
		{
		    destination[dest_count++] = 0x1B;
		    if(dest_count<max)
			destination[dest_count++] = ext_charset[table_row*2+1];
		    found = 1;
		    break;
		}
		table_row++;
	    }
	}
	
	if (found==0)
	    QMessageBox::warning(this, "SMS-Tool", QString("The character %1 could not be translated into the GSM character set and will be ignored therefore.").arg(source[source_count]), QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton, QMessageBox::NoButton);
	
	source_count++;
    }
    
    // terminate destination string with 0, however 0x00 are also allowed within the string
    destination[dest_count] = 0;
    return dest_count;
}

bool SMSTool::send_part(int modem, const char *number, char *message, int messagelen, const char *udh, QProgressDialog &progress)
{
    int coding, flags, numberformat, numberlength;
    char tmp[50], tmp2[500], pdu[600], command[650], answer[500];
    
    if(number[0] == '+')
    {
	numberformat = 145;
	strcpy(tmp, number+1);
    }
    else
    {
	numberformat = 129;
	strcpy(tmp, number);
    }
    
    numberlength = strlen(tmp);
    
    // terminate the number with F if the length is odd
    if(numberlength % 2)
	strcat(tmp,"F");
    
    // swap every second character
    swapchars(tmp);
    
    flags=1; // use SMSC
    flags+=16; // validity field
    if(udh[0] != 0)
	flags += 64; // user data header (UDH)
    
    coding=240; // 7-bit text
    coding+=1; // class 1 (no flash SMS)
    
    // make PDU
    if(udh[0] == 0)
    {
	messagelen = ascii2pdu(message, messagelen, tmp2);
	// http://www.dreamfabric.com/sms/
	sprintf(pdu, "00%02X00%02X%02X%s00%02XFF%02X", flags, numberlength, numberformat, tmp, coding, messagelen);
    }
    else
    {
	messagelen = ascii2pdu(message+1, messagelen-1, tmp2);
	messagelen += 8; // 6 bytes UDH + 1 byte (with a fill bit and the first character) to get to the next 7-boundary => 8 7-bit characters
	sprintf(pdu, "00%02X00%02X%02X%s00%02XFF%02X%s%02X", flags, numberlength, numberformat, tmp, coding, messagelen, udh, ((unsigned char)message[0]) << 1);
    }
    strcat(pdu, tmp2);
    
    int offset = progress.progress();
    
    // send the message
    sprintf(command, "AT+CMGS=%i\r", strlen(pdu) / 2 - 1);
//    printf("command: %s\n", command);
    modem_command(modem, command, answer, sizeof(answer), 50, "(\\>)|(ERROR)", NULL);
//    printf("answer: %s\n", answer);
    if(strstr(answer, "ERROR"))
	return false;
    
    progress.setProgress(offset + 100);
    
    sprintf(command, "%s\x1A", pdu);
//    printf("command: %s\n", command);
    modem_command(modem, command, answer, sizeof(answer), 300, "(OK)|(ERROR)", &progress);
//    printf("answer: %s\n", answer);
    if(strstr(answer, "ERROR"))
	return false;
    
    return true;
}
